﻿# Define the API URIs
param(
  [parameter(Position=0,Mandatory=$false)][boolean] $useHttps = $true,
  [parameter(Position=1,Mandatory=$false)][boolean] $ValidateFQDN = $true
)

[string] $EzUdmAuthApi = "api/ps/auth"
[string] $EzUdmLocationApi = "api/ps/location"
[string] $EzUdmRmRangeApi = "api/ps/range"
[string] $EzUdmRmNumberBlockApi = "api/ps/numberBlock"
[string] $EzUdmRmNumberApi = "api/ps/number"
[string] $EzUdmRmNumberReservationApi = "api/ps/numberReservation"
[string] $HelpUriRoot = 'https://eventzero.atlassian.net/wiki/pages/viewpage.action?pageId='
[string] $HTTP = if($useHttps){'https://'} else {'http://'}

### constants
[string] $EMAIL_FORMAT = '^[A-Z0-9._%+-]+@[A-Z0-9.-]+\.[A-Z]{2,}$'
[string] $SIP_FORMAT = '^[a-zA-Z0-9._%+-]+@[a-zA-Z0-9.-]+\.[a-zA-Z]{2,}$'
[string] $PHONE_NUMBER_FORMAT = '^\+?[0-9]{1,40}$'
[string] $EXTENSION_FORMAT = '^[0-9]{0,40}$'
[string] $QUANTITY_FORMAT = '^[1-9][0-9]{0,5}$'
[string] $TIME_FORMAT = '^[0-9]{0,7}$'

function Connect-UdmServer{
  <#
      .SYNOPSIS
      Connects to UDM Server and store the crendentials for later use.

      .DESCRIPTION
      Connects to UDM Server and store the crendentials for later use.

      .EXAMPLE
      $cred = Get-Credential
      Connect-UdmServer -Credentials $cred -Tenant contoso.udm.eventzero.net
      Connects you to the contoso.udm.eventzero.net tenant using the credentials supplied to the get-credentials command

      .NOTES
      Version 1.0

      .INPUTS
      List of input types that are accepted by this function.

      .OUTPUTS
      None. Connect-UdmServer does not generate any output.      
  #>


  [CmdletBinding(SupportsShouldProcess, SupportsPaging, HelpURI = '')]
  param(
    [Parameter(ValueFromPipelineByPropertyName)]
    [System.Management.Automation.Credential()][PSCredential] $Credentials,
 
    [Parameter(ValueFromPipeline)]
    [ValidateScript({If (!$ValidateFQDN -or $_ -match '(?=^.{4,253}$)(^((?!-)[a-zA-Z0-9-]{1,63}(?<!-)\.)+[a-zA-Z]{2,63}$)') {
      $True
    } else {
      Throw "$_ is not in FQDN format!"
    }})]
    [string] $Tenant
  )
  Begin{
    [string] $TenantSourceXml = ".\EzUdmTenants.xml"
  }
  Process{
    try{
      # check for supplied tenant
      if ((-not ($Tenant)) -and (-not ($global:EzUdmTenant))){
        if (Test-Path -Path $TenantSourceXml){
          [xml] $xml = Get-Content -Path $TenantSourceXml
          if ($xml){
            $AllTenants = $xml.tenants.tenant.fqdn | Sort-Object -Property id
            if ($AllTenants.count -gt 1){
              $Tenant = ($AllTenants | Out-GridView -Title 'Choose a tenant to connect to' -OutputMode Single).id
            }else{
              $Tenant = ($AllTenants).id
            }
          }
        }else{
          $Tenant = Read-Host -Prompt 'Enter tenant name'
        }
        if ($ValidateFQDN -and $Tenant -notmatch '(?=^.{4,253}$)(^((?!-)[a-zA-Z0-9-]{1,63}(?<!-)\.)+[a-zA-Z]{2,63}$)'){
          Write-Output -InputObject ('Tenant "{0}" supplied is not an FQDN' -f $Tenant)
          return
        }
      }
      
      # check for stored encrypted credentials
      # if multiple creds found, show in Out-Grid for user to pick
      # prompt for credentials if not found
      if ((-not ($Credentials)) -and (-not ($global:EzUdmCreds))){
          $Credentials = Get-Credential
      }

      if (-not $Tenant) {
        $Tenant = $global:EzUdmTenant
      }
      if (-not $Credentials) {
        $Credentials = $global:EzUdmCreds
      }

      $Uri = $HTTP + $Tenant + '/' + $EzUdmAuthApi
      # What-If suport
      if (-not $PSCmdlet.ShouldProcess($Uri)) {
        return
      }

      # check for connection to tenant
      $WebRequest = Invoke-WebRequest -Uri $Uri -Method Get -Credential $Credentials
      if ($WebRequest.StatusCode -ne 200) {
        Write-Output -InputObject ('Non-200 response from connect: {0}' -f $WebRequest.StatusCode)
      } elseif ($WebRequest.Content -ne "OK") {
        Write-Output -InputObject ('Expected response to be "OK" but it was {0}' -f $WebRequest.Content)
      }
    
      # store credentials in global variable $global:EzUdmCreds - this needs to be a global variable so that it survives for the duration of the PS session
      # store tenant info - this needs to be a global variable so that it survives for the duration of the PS session
      $global:EzUdmCreds = $Credentials
      $global:EzUdmTenant = $Tenant
    }
    catch{
      Throw "An error has occured. The error was: '$_'."
    }
  }
  End{
    # connected - Boolean value if successful
    Write-Output -InputObject $global:EzUdmTenant
    # Remove-Variable -Name EzUdmTenant -Scope global
    # Remove-Variable -Name Tenant
  }
} # end function Connect-UdmServer

#region UdmLocation
function New-UdmLocation{
  <#
      .SYNOPSIS
      Creates a new Location within the core UDM Provisioning product. Note it does not create 

      .DESCRIPTION
      Creates a new Location within the core UDM Provisioning product. 

      .EXAMPLE
      New-UdmLocation -Name Value -Notes Value -Country Value
      Describe what this call does

      .NOTES
      Version 1.0

      .INPUTS
      List of input types that are accepted by this function.

      .OUTPUTS
      system.object
  #>


  [CmdletBinding(SupportsShouldProcess, SupportsPaging, HelpURI = '')]
  param(
    # Name of the new Location
    [Parameter(Mandatory=$true, ValueFromPipeline)]
    [ValidateNotNullOrEmpty()]
    [ValidateLength(1,40)]
    [string] $Name,
    
    # Freeform notes to assign to this Location
    [Parameter(ValueFromPipelineByPropertyName)]
    [ValidateLength(1,255)]
    [string] $Note,

    # First and last name of the contact person for this location
    [Parameter(ValueFromPipelineByPropertyName)]
    [string] $ContactName,
    
    # Phone number of the contact person for this location
    [Parameter(ValueFromPipelineByPropertyName)]
    [string] $ContactNumber,
    
    # Email address of the contact person for this location
    [Parameter(ValueFromPipelineByPropertyName)]
    [AllowNull()]
    [ValidateScript({If ($_ -imatch $EMAIL_FORMAT) {
      $True
    } Else {
      Throw "Input value '$_' is not in SMTP format!"
    }})]    
    [string] $ContactEmail,

    # Street address of the location
    [Parameter(ValueFromPipelineByPropertyName)]
    [string] $Address,
    
    # City that the location is in
    [Parameter(ValueFromPipelineByPropertyName)]
    [string] $City,
    
    # State that the location is in
    [Parameter(ValueFromPipelineByPropertyName)]
    [string] $State,
    
    # Postal code location is in
    [Parameter(ValueFromPipelineByPropertyName)]
    [string] $PostalCode,
    
    # Country that the new Location is in. Acceptable values are based on the ISO list of countries.
    [Parameter(ValueFromPipelineByPropertyName)]
    [ValidateLength(1,44)]
    [ValidateSet('Afghanistan', 'Aland Islands', 'Albania', 'Algeria', 'American Samoa', 'Andorra', 'Angola', 'Anguilla', 'Antarctica', 'Antigua and Barbuda', 'Argentina', 'Armenia', 'Aruba', 'Australia', 'Austria', 'Azerbaijan', 'Bahamas', 'Bahrain', 'Bangladesh', 'Barbados', 'Belarus', 'Belgium', 'Belize', 'Benin', 'Bermuda', 'Bhutan', 'Bolivia, Plurinational State of', 'Bonaire, Sint Eustatius and Saba', 'Bosnia and Herzegovina', 'Botswana', 'Bouvet Island', 'Brazil', 'British Indian Ocean Territory', 'Brunei Darussalam', 'Bulgaria', 'Burkina Faso', 'Burundi', 'Cambodia', 'Cameroon', 'Canada', 'Cape Verde', 'Cayman Islands', 'Central African Republic', 'Chad', 'Chile', 'China', 'Christmas Island', 'Cocos (Keeling) Islands', 'Colombia', 'Comoros', 'Congo', 'Congo, the Democratic Republic of the', 'Cook Islands', 'Costa Rica', "Cote d'Ivoire", 'Croatia', 'Cuba', 'Curacao', 'Cyprus', 'Czech Republic', 'Denmark', 'Djibouti', 'Dominica', 'Dominican Republic', 'Ecuador', 'Egypt', 'El Salvador', 'Equatorial Guinea', 'Eritrea', 'Estonia', 'Ethiopia', 'Falkland Islands (Malvinas)', 'Faroe Islands', 'Fiji', 'Finland', 'France', 'French Guiana', 'French Polynesia', 'French Southern Territories', 'Gabon', 'Gambia', 'Georgia', 'Germany', 'Ghana', 'Gibraltar', 'Greece', 'Greenland', 'Grenada', 'Guadeloupe', 'Guam', 'Guatemala', 'Guernsey', 'Guinea', 'Guinea-Bissau', 'Guyana', 'Haiti', 'Heard Island and McDonald Islands', 'Holy See (Vatican City State)', 'Honduras', 'Hong Kong', 'Hungary', 'Iceland', 'India', 'Indonesia', 'Iran, Islamic Republic of', 'Iraq', 'Ireland', 'Isle of Man', 'Israel', 'Italy', 'Jamaica', 'Japan', 'Jersey', 'Jordan', 'Kazakhstan', 'Kenya', 'Kiribati', "Korea, Democratic People's Republic of", 'Korea, Republic of', 'Kuwait', 'Kyrgyzstan', "Lao People's Democratic Republic", 'Latvia', 'Lebanon', 'Lesotho', 'Liberia', 'Libya', 'Liechtenstein', 'Lithuania', 'Luxembourg', 'Macao', 'Macedonia, the former Yugoslav Republic of', 'Madagascar', 'Malawi', 'Malaysia', 'Maldives', 'Mali', 'Malta', 'Marshall Islands', 'Martinique', 'Mauritania', 'Mauritius', 'Mayotte', 'Mexico', 'Micronesia, Federated States of', 'Moldova, Republic of', 'Monaco', 'Mongolia', 'Montenegro', 'Montserrat', 'Morocco', 'Mozambique', 'Myanmar', 'Namibia', 'Nauru', 'Nepal', 'Netherlands', 'New Caledonia', 'New Zealand', 'Nicaragua', 'Niger', 'Nigeria', 'Niue', 'Norfolk Island', 'Northern Mariana Islands', 'Norway', 'Oman', 'Pakistan', 'Palau', 'Palestine, State of', 'Panama', 'Papua New Guinea', 'Paraguay', 'Peru', 'Philippines', 'Pitcairn', 'Poland', 'Portugal', 'Puerto Rico', 'Qatar', 'Reunion', 'Romania', 'Russian Federation', 'Rwanda', 'Saint Barthelemy', 'Saint Helena, Ascension and Tristan da Cunha', 'Saint Kitts and Nevis', 'Saint Lucia', 'Saint Martin (French part)', 'Saint Pierre and Miquelon', 'Saint Vincent and the Grenadines', 'Samoa', 'San Marino', 'Sao Tome and Principe', 'Saudi Arabia', 'Senegal', 'Serbia', 'Seychelles', 'Sierra Leone', 'Singapore', 'Sint Maarten (Dutch part)', 'Slovakia', 'Slovenia', 'Solomon Islands', 'Somalia', 'South Africa', 'South Georgia and the South Sandwich Islands', 'South Sudan', 'Spain', 'Sri Lanka', 'Sudan', 'Suriname', 'Svalbard and Jan Mayen', 'Swaziland', 'Sweden', 'Switzerland', 'Syrian Arab Republic', 'Taiwan, Province of China', 'Tajikistan', 'Tanzania, United Republic of', 'Thailand', 'Timor-Leste', 'Togo', 'Tokelau', 'Tonga', 'Trinidad and Tobago', 'Tunisia', 'Turkey', 'Turkmenistan', 'Turks and Caicos Islands', 'Tuvalu', 'Uganda', 'Ukraine', 'United Arab Emirates', 'United Kingdom', 'United States', 'United States Minor Outlying Islands', 'Uruguay', 'Uzbekistan', 'Vanuatu', 'Venezuela, Bolivarian Republic of', 'Viet Nam', 'Virgin Islands, British', 'Virgin Islands, U.S.', 'Wallis and Futuna', 'Western Sahara', 'Yemen', 'Zambia', 'Zimbabwe')]
    [string] $Country,
    
    # Timezone for the new location. Acceptable values are based on the ISO list of timezones.
    [Parameter(ValueFromPipelineByPropertyName)]
    [ValidateLength(1, 32)]
    [ValidateSet('ACT', 'AET', 'AGT', 'ART', 'AST', 'Africa/Abidjan', 'Africa/Accra', 'Africa/Addis_Ababa', 'Africa/Algiers', 'Africa/Asmara', 'Africa/Asmera', 'Africa/Bamako', 'Africa/Bangui', 'Africa/Banjul', 'Africa/Bissau', 'Africa/Blantyre', 'Africa/Brazzaville', 'Africa/Bujumbura', 'Africa/Cairo', 'Africa/Casablanca', 'Africa/Ceuta', 'Africa/Conakry', 'Africa/Dakar', 'Africa/Dar_es_Salaam', 'Africa/Djibouti', 'Africa/Douala', 'Africa/El_Aaiun', 'Africa/Freetown', 'Africa/Gaborone', 'Africa/Harare', 'Africa/Johannesburg', 'Africa/Juba', 'Africa/Kampala', 'Africa/Khartoum', 'Africa/Kigali', 'Africa/Kinshasa', 'Africa/Lagos', 'Africa/Libreville', 'Africa/Lome', 'Africa/Luanda', 'Africa/Lubumbashi', 'Africa/Lusaka', 'Africa/Malabo', 'Africa/Maputo', 'Africa/Maseru', 'Africa/Mbabane', 'Africa/Mogadishu', 'Africa/Monrovia', 'Africa/Nairobi', 'Africa/Ndjamena', 'Africa/Niamey', 'Africa/Nouakchott', 'Africa/Ouagadougou', 'Africa/Porto-Novo', 'Africa/Sao_Tome', 'Africa/Timbuktu', 'Africa/Tripoli', 'Africa/Tunis', 'Africa/Windhoek', 'America/Adak', 'America/Anchorage', 'America/Anguilla', 'America/Antigua', 'America/Araguaina', 'America/Argentina/Buenos_Aires', 'America/Argentina/Catamarca', 'America/Argentina/ComodRivadavia', 'America/Argentina/Cordoba', 'America/Argentina/Jujuy', 'America/Argentina/La_Rioja', 'America/Argentina/Mendoza', 'America/Argentina/Rio_Gallegos', 'America/Argentina/Salta', 'America/Argentina/San_Juan', 'America/Argentina/San_Luis', 'America/Argentina/Tucuman', 'America/Argentina/Ushuaia', 'America/Aruba', 'America/Asuncion', 'America/Atikokan', 'America/Atka', 'America/Bahia', 'America/Bahia_Banderas', 'America/Barbados', 'America/Belem', 'America/Belize', 'America/Blanc-Sablon', 'America/Boa_Vista', 'America/Bogota', 'America/Boise', 'America/Buenos_Aires', 'America/Cambridge_Bay', 'America/Campo_Grande', 'America/Cancun', 'America/Caracas', 'America/Catamarca', 'America/Cayenne', 'America/Cayman', 'America/Chicago', 'America/Chihuahua', 'America/Coral_Harbour', 'America/Cordoba', 'America/Costa_Rica', 'America/Creston', 'America/Cuiaba', 'America/Curacao', 'America/Danmarkshavn', 'America/Dawson', 'America/Dawson_Creek', 'America/Denver', 'America/Detroit', 'America/Dominica', 'America/Edmonton', 'America/Eirunepe', 'America/El_Salvador', 'America/Ensenada', 'America/Fort_Wayne', 'America/Fortaleza', 'America/Glace_Bay', 'America/Godthab', 'America/Goose_Bay', 'America/Grand_Turk', 'America/Grenada', 'America/Guadeloupe', 'America/Guatemala', 'America/Guayaquil', 'America/Guyana', 'America/Halifax', 'America/Havana', 'America/Hermosillo', 'America/Indiana/Indianapolis', 'America/Indiana/Knox', 'America/Indiana/Marengo', 'America/Indiana/Petersburg', 'America/Indiana/Tell_City', 'America/Indiana/Vevay', 'America/Indiana/Vincennes', 'America/Indiana/Winamac', 'America/Indianapolis', 'America/Inuvik', 'America/Iqaluit', 'America/Jamaica', 'America/Jujuy', 'America/Juneau', 'America/Kentucky/Louisville', 'America/Kentucky/Monticello', 'America/Knox_IN', 'America/Kralendijk', 'America/La_Paz', 'America/Lima', 'America/Los_Angeles', 'America/Louisville', 'America/Lower_Princes', 'America/Maceio', 'America/Managua', 'America/Manaus', 'America/Marigot', 'America/Martinique', 'America/Matamoros', 'America/Mazatlan', 'America/Mendoza', 'America/Menominee', 'America/Merida', 'America/Metlakatla', 'America/Mexico_City', 'America/Miquelon', 'America/Moncton', 'America/Monterrey', 'America/Montevideo', 'America/Montreal', 'America/Montserrat', 'America/Nassau', 'America/New_York', 'America/Nipigon', 'America/Nome', 'America/Noronha', 'America/North_Dakota/Beulah', 'America/North_Dakota/Center', 'America/North_Dakota/New_Salem', 'America/Ojinaga', 'America/Panama', 'America/Pangnirtung', 'America/Paramaribo', 'America/Phoenix', 'America/Port-au-Prince', 'America/Port_of_Spain', 'America/Porto_Acre', 'America/Porto_Velho', 'America/Puerto_Rico', 'America/Rainy_River', 'America/Rankin_Inlet', 'America/Recife', 'America/Regina', 'America/Resolute', 'America/Rio_Branco', 'America/Rosario', 'America/Santa_Isabel', 'America/Santarem', 'America/Santiago', 'America/Santo_Domingo', 'America/Sao_Paulo', 'America/Scoresbysund', 'America/Shiprock', 'America/Sitka', 'America/St_Barthelemy', 'America/St_Johns', 'America/St_Kitts', 'America/St_Lucia', 'America/St_Thomas', 'America/St_Vincent', 'America/Swift_Current', 'America/Tegucigalpa', 'America/Thule', 'America/Thunder_Bay', 'America/Tijuana', 'America/Toronto', 'America/Tortola', 'America/Vancouver', 'America/Virgin', 'America/Whitehorse', 'America/Winnipeg', 'America/Yakutat', 'America/Yellowknife', 'Antarctica/Casey', 'Antarctica/Davis', 'Antarctica/DumontDUrville', 'Antarctica/Macquarie', 'Antarctica/Mawson', 'Antarctica/McMurdo', 'Antarctica/Palmer', 'Antarctica/Rothera', 'Antarctica/South_Pole', 'Antarctica/Syowa', 'Antarctica/Vostok', 'Arctic/Longyearbyen', 'Asia/Aden', 'Asia/Almaty', 'Asia/Amman', 'Asia/Anadyr', 'Asia/Aqtau', 'Asia/Aqtobe', 'Asia/Ashgabat', 'Asia/Ashkhabad', 'Asia/Baghdad', 'Asia/Bahrain', 'Asia/Baku', 'Asia/Bangkok', 'Asia/Beirut', 'Asia/Bishkek', 'Asia/Brunei', 'Asia/Calcutta', 'Asia/Choibalsan', 'Asia/Chongqing', 'Asia/Chungking', 'Asia/Colombo', 'Asia/Dacca', 'Asia/Damascus', 'Asia/Dhaka', 'Asia/Dili', 'Asia/Dubai', 'Asia/Dushanbe', 'Asia/Gaza', 'Asia/Harbin', 'Asia/Hebron', 'Asia/Ho_Chi_Minh', 'Asia/Hong_Kong', 'Asia/Hovd', 'Asia/Irkutsk', 'Asia/Istanbul', 'Asia/Jakarta', 'Asia/Jayapura', 'Asia/Jerusalem', 'Asia/Kabul', 'Asia/Kamchatka', 'Asia/Karachi', 'Asia/Kashgar', 'Asia/Kathmandu', 'Asia/Katmandu', 'Asia/Kolkata', 'Asia/Krasnoyarsk', 'Asia/Kuala_Lumpur', 'Asia/Kuching', 'Asia/Kuwait', 'Asia/Macao', 'Asia/Macau', 'Asia/Magadan', 'Asia/Makassar', 'Asia/Manila', 'Asia/Muscat', 'Asia/Nicosia', 'Asia/Novokuznetsk', 'Asia/Novosibirsk', 'Asia/Omsk', 'Asia/Oral', 'Asia/Phnom_Penh', 'Asia/Pontianak', 'Asia/Pyongyang', 'Asia/Qatar', 'Asia/Qyzylorda', 'Asia/Rangoon', 'Asia/Riyadh', 'Asia/Riyadh87', 'Asia/Riyadh88', 'Asia/Riyadh89', 'Asia/Saigon', 'Asia/Sakhalin', 'Asia/Samarkand', 'Asia/Seoul', 'Asia/Shanghai', 'Asia/Singapore', 'Asia/Taipei', 'Asia/Tashkent', 'Asia/Tbilisi', 'Asia/Tehran', 'Asia/Tel_Aviv', 'Asia/Thimbu', 'Asia/Thimphu', 'Asia/Tokyo', 'Asia/Ujung_Pandang', 'Asia/Ulaanbaatar', 'Asia/Ulan_Bator', 'Asia/Urumqi', 'Asia/Vientiane', 'Asia/Vladivostok', 'Asia/Yakutsk', 'Asia/Yekaterinburg', 'Asia/Yerevan', 'Atlantic/Azores', 'Atlantic/Bermuda', 'Atlantic/Canary', 'Atlantic/Cape_Verde', 'Atlantic/Faeroe', 'Atlantic/Faroe', 'Atlantic/Jan_Mayen', 'Atlantic/Madeira', 'Atlantic/Reykjavik', 'Atlantic/South_Georgia', 'Atlantic/St_Helena', 'Atlantic/Stanley', 'Australia/ACT', 'Australia/Adelaide', 'Australia/Brisbane', 'Australia/Broken_Hill', 'Australia/Canberra', 'Australia/Currie', 'Australia/Darwin', 'Australia/Eucla', 'Australia/Hobart', 'Australia/LHI', 'Australia/Lindeman', 'Australia/Lord_Howe', 'Australia/Melbourne', 'Australia/NSW', 'Australia/North', 'Australia/Perth', 'Australia/Queensland', 'Australia/South', 'Australia/Sydney', 'Australia/Tasmania', 'Australia/Victoria', 'Australia/West', 'Australia/Yancowinna', 'BET', 'BST', 'Brazil/Acre', 'Brazil/DeNoronha', 'Brazil/East', 'Brazil/West', 'CAT', 'CET', 'CNT', 'CST', 'CST6CDT', 'CTT', 'Canada/Atlantic', 'Canada/Central', 'Canada/East-Saskatchewan', 'Canada/Eastern', 'Canada/Mountain', 'Canada/Newfoundland', 'Canada/Pacific', 'Canada/Saskatchewan', 'Canada/Yukon', 'Chile/Continental', 'Chile/EasterIsland', 'Cuba', 'EAT', 'ECT', 'EET', 'EST', 'EST5EDT', 'Egypt', 'Eire', 'Etc/GMT', 'Etc/GMT+0', 'Etc/GMT+1', 'Etc/GMT+10', 'Etc/GMT+11', 'Etc/GMT+12', 'Etc/GMT+2', 'Etc/GMT+3', 'Etc/GMT+4', 'Etc/GMT+5', 'Etc/GMT+6', 'Etc/GMT+7', 'Etc/GMT+8', 'Etc/GMT+9', 'Etc/GMT-0', 'Etc/GMT-1', 'Etc/GMT-10', 'Etc/GMT-11', 'Etc/GMT-12', 'Etc/GMT-13', 'Etc/GMT-14', 'Etc/GMT-2', 'Etc/GMT-3', 'Etc/GMT-4', 'Etc/GMT-5', 'Etc/GMT-6', 'Etc/GMT-7', 'Etc/GMT-8', 'Etc/GMT-9', 'Etc/GMT0', 'Etc/Greenwich', 'Etc/UCT', 'Etc/UTC', 'Etc/Universal', 'Etc/Zulu', 'Europe/Amsterdam', 'Europe/Andorra', 'Europe/Athens', 'Europe/Belfast', 'Europe/Belgrade', 'Europe/Berlin', 'Europe/Bratislava', 'Europe/Brussels', 'Europe/Bucharest', 'Europe/Budapest', 'Europe/Chisinau', 'Europe/Copenhagen', 'Europe/Dublin', 'Europe/Gibraltar', 'Europe/Guernsey', 'Europe/Helsinki', 'Europe/Isle_of_Man', 'Europe/Istanbul', 'Europe/Jersey', 'Europe/Kaliningrad', 'Europe/Kiev', 'Europe/Lisbon', 'Europe/Ljubljana', 'Europe/London', 'Europe/Luxembourg', 'Europe/Madrid', 'Europe/Malta', 'Europe/Mariehamn', 'Europe/Minsk', 'Europe/Monaco', 'Europe/Moscow', 'Europe/Nicosia', 'Europe/Oslo', 'Europe/Paris', 'Europe/Podgorica', 'Europe/Prague', 'Europe/Riga', 'Europe/Rome', 'Europe/Samara', 'Europe/San_Marino', 'Europe/Sarajevo', 'Europe/Simferopol', 'Europe/Skopje', 'Europe/Sofia', 'Europe/Stockholm', 'Europe/Tallinn', 'Europe/Tirane', 'Europe/Tiraspol', 'Europe/Uzhgorod', 'Europe/Vaduz', 'Europe/Vatican', 'Europe/Vienna', 'Europe/Vilnius', 'Europe/Volgograd', 'Europe/Warsaw', 'Europe/Zagreb', 'Europe/Zaporozhye', 'Europe/Zurich', 'Factory', 'GB', 'GB-Eire', 'GMT', 'GMT+0', 'GMT-0', 'GMT0', 'Greenwich', 'HST', 'Hongkong', 'IET', 'IST', 'Iceland', 'Indian/Antananarivo', 'Indian/Chagos', 'Indian/Christmas', 'Indian/Cocos', 'Indian/Comoro', 'Indian/Kerguelen', 'Indian/Mahe', 'Indian/Maldives', 'Indian/Mauritius', 'Indian/Mayotte', 'Indian/Reunion', 'Iran', 'Israel', 'JST', 'Jamaica', 'Japan', 'Kwajalein', 'Libya', 'MET', 'MIT', 'MST', 'MST7MDT', 'Mexico/BajaNorte', 'Mexico/BajaSur', 'Mexico/General', 'Mideast/Riyadh87', 'Mideast/Riyadh88', 'Mideast/Riyadh89', 'NET', 'NST', 'NZ', 'NZ-HAT', 'Navajo', 'PLT', 'PNT', 'PRC', 'PRT', 'PST', 'PST8PDT', 'Pacific/Apia', 'Pacific/Auckland', 'Pacific/Chatham', 'Pacific/Chuuk', 'Pacific/Easter', 'Pacific/Efate', 'Pacific/Enderbury', 'Pacific/Fakaofo', 'Pacific/Fiji', 'Pacific/Funafuti', 'Pacific/Galapagos', 'Pacific/Gambier', 'Pacific/Guadalcanal', 'Pacific/Guam', 'Pacific/Honolulu', 'Pacific/Johnston', 'Pacific/Kiritimati', 'Pacific/Kosrae', 'Pacific/Kwajalein', 'Pacific/Majuro', 'Pacific/Marquesas', 'Pacific/Midway', 'Pacific/Nauru', 'Pacific/Niue', 'Pacific/Norfolk', 'Pacific/Noumea', 'Pacific/Pago_Pago', 'Pacific/Palau', 'Pacific/Pitcairn', 'Pacific/Pohnpei', 'Pacific/Ponape', 'Pacific/Port_Moresby', 'Pacific/Rarotonga', 'Pacific/Saipan', 'Pacific/Samoa', 'Pacific/Tahiti', 'Pacific/Tarawa', 'Pacific/Tongatapu', 'Pacific/Truk', 'Pacific/Wake', 'Pacific/Wallis', 'Pacific/Yap', 'Poland', 'Portugal', 'ROC', 'ROK', 'SST', 'Singapore', 'SystemV/AST4', 'SystemV/AST4ADT', 'SystemV/CST6', 'SystemV/CST6CDT', 'SystemV/EST5', 'SystemV/EST5EDT', 'SystemV/HST10', 'SystemV/MST7', 'SystemV/MST7MDT', 'SystemV/PST8', 'SystemV/PST8PDT', 'SystemV/YST9', 'SystemV/YST9YDT', 'Turkey', 'UCT', 'US/Alaska', 'US/Aleutian', 'US/Arizona', 'US/Central', 'US/East-Indiana', 'US/Eastern', 'US/Hawaii', 'US/Indiana-Starke', 'US/Michigan', 'US/Mountain', 'US/Pacific', 'US/Pacific-New', 'US/Samoa', 'UTC', 'Universal', 'VST', 'W-SU', 'WET', 'Zulu')]
    [string] $TimeZone
  )
  Begin{
    ConnectToServerIfNeeded
  }
  Process{
    try{
      $Params = @{}
      if ($Name){$Params["Name"] = $Name}
      if ($Note){$Params["Note"] = $Note}
      if ($ContactName){$Params["ContactName"] = $ContactName}
      if ($ContactNumber){$Params["ContactNumber"] = $ContactNumber}
      if ($ContactEmail){$Params["ContactEmail"] = $ContactEmail}
      if ($Address){$Params["Address"] = $Address}
      if ($City){$Params["City"] = $City}
      if ($State){$Params["State"] = $State}
      if ($PostalCode){$Params["PostalCode"] = $PostalCode}
      if ($Country){$Params["Country"] = $Country}
      if ($TimeZone){$Params["TimeZone"] = $TimeZone}

      # Assemble the URI
      [string] $Uri = $HTTP + $global:EzUdmTenant + '/' + $EzUdmLocationApi
      Write-Verbose -Message $Uri

      # What-If suport
      if (-not $(ShouldProcess $Params)) {
        return
      }

      $WebRequest = Invoke-WebRequest -Uri $Uri -Method Post -Credential $global:EzUdmCreds -Body $Params
    
      # Handle the JSON data returned
      # This code is just working theory right now, and should not be trusted
      $parsed = ConvertFrom-Json -InputObject $WebRequest.Content
      return $parsed
    }
    catch{
      Throw "An error has occured. The error was: '$_'."
    }
  }
  End{}
} # end function New-UdmLocation
function Set-UdmLocation{
  <#
      .SYNOPSIS
      Updates an existing Location within the core UDM product.

      .DESCRIPTION
      Updates an existing Location within the core UDM product.

      .EXAMPLE
      Set-UdmLocation -Identity Value -Notes Value -Country Value
      Describe what this call does

      .NOTES
      Version 1.0

      .INPUTS
      List of input types that are accepted by this function.

      .OUTPUTS
      system.object
  #>


  [CmdletBinding(SupportsShouldProcess, SupportsPaging, HelpURI = '')]
  param(
    # Identity of the Location to be changed
    [Parameter(Mandatory=$true, ValueFromPipelineByPropertyName)]
    [Alias('id')]
    [ValidateNotNullOrEmpty()]
    [ValidateLength(1,40)]
    [string] $Identity,

    [Parameter(ValueFromPipelineByPropertyName)]
    [ValidateNotNullOrEmpty()]
    [ValidateLength(1,40)]
    [string] $Name, 
    
    # Freeform notes to assign to this 
    [Parameter(ValueFromPipelineByPropertyName)]
    [ValidateLength(1,255)]
    [string] $Note,
    
    # First and last name of the contact person for this location
    [Parameter(ValueFromPipelineByPropertyName)]
    [string] $ContactName,
    
    # Phone number of the contact person for this location
    [Parameter(ValueFromPipelineByPropertyName)]
    [string] $ContactNumber,
    
    # Email address of the contact person for this location
    [Parameter(ValueFromPipelineByPropertyName)]
    [AllowNull()]
    [AllowEmptyString()]
    [ValidateScript({If ($_ -eq '' -or $_ -imatch $EMAIL_FORMAT) {
      $True
    } Else {
      Throw "Input '$_' is not in SMTP format!"
    }})]    
    [string] $ContactEmail,
    
    # SIP address of the contact person for this location
    [Parameter(ValueFromPipelineByPropertyName)]
    [ValidateScript({
      If ($_ -eq '' -or $_ -match $SIP_FORMAT) {
      $True
    } Else {
      Throw "'$_' is not in SIP format!"
    }})]
    [string] $ContactSipAddress,
    
    # Street address of the location
    [Parameter(ValueFromPipelineByPropertyName)]
    [string] $Address,
    
    # city that the location is in
    [Parameter(ValueFromPipelineByPropertyName)]
    [string] $City,
    
    # State that the location is in
    [Parameter(ValueFromPipelineByPropertyName)]
    [string] $State,
    
    # Postal code location is in
    [Parameter(ValueFromPipelineByPropertyName)]
    [string] $PostalCode,
    
    # New Country that the Location is in. Acceptable values are based on the ISO list of countries.
    [Parameter(ValueFromPipelineByPropertyName)]
    [AllowNull()]
    [AllowEmptyString()]
    [ValidateLength(0,44)]
    [ValidateSet('Afghanistan', 'Aland Islands', 'Albania', 'Algeria', 'American Samoa', 'Andorra', 'Angola', 'Anguilla', 'Antarctica', 'Antigua and Barbuda', 'Argentina', 'Armenia', 'Aruba', 'Australia', 'Austria', 'Azerbaijan', 'Bahamas', 'Bahrain', 'Bangladesh', 'Barbados', 'Belarus', 'Belgium', 'Belize', 'Benin', 'Bermuda', 'Bhutan', 'Bolivia, Plurinational State of', 'Bonaire, Sint Eustatius and Saba', 'Bosnia and Herzegovina', 'Botswana', 'Bouvet Island', 'Brazil', 'British Indian Ocean Territory', 'Brunei Darussalam', 'Bulgaria', 'Burkina Faso', 'Burundi', 'Cambodia', 'Cameroon', 'Canada', 'Cape Verde', 'Cayman Islands', 'Central African Republic', 'Chad', 'Chile', 'China', 'Christmas Island', 'Cocos (Keeling) Islands', 'Colombia', 'Comoros', 'Congo', 'Congo, the Democratic Republic of the', 'Cook Islands', 'Costa Rica', "Cote d'Ivoire", 'Croatia', 'Cuba', 'Curacao', 'Cyprus', 'Czech Republic', 'Denmark', 'Djibouti', 'Dominica', 'Dominican Republic', 'Ecuador', 'Egypt', 'El Salvador', 'Equatorial Guinea', 'Eritrea', 'Estonia', 'Ethiopia', 'Falkland Islands (Malvinas)', 'Faroe Islands', 'Fiji', 'Finland', 'France', 'French Guiana', 'French Polynesia', 'French Southern Territories', 'Gabon', 'Gambia', 'Georgia', 'Germany', 'Ghana', 'Gibraltar', 'Greece', 'Greenland', 'Grenada', 'Guadeloupe', 'Guam', 'Guatemala', 'Guernsey', 'Guinea', 'Guinea-Bissau', 'Guyana', 'Haiti', 'Heard Island and McDonald Islands', 'Holy See (Vatican City State)', 'Honduras', 'Hong Kong', 'Hungary', 'Iceland', 'India', 'Indonesia', 'Iran, Islamic Republic of', 'Iraq', 'Ireland', 'Isle of Man', 'Israel', 'Italy', 'Jamaica', 'Japan', 'Jersey', 'Jordan', 'Kazakhstan', 'Kenya', 'Kiribati', "Korea, Democratic People's Republic of", 'Korea, Republic of', 'Kuwait', 'Kyrgyzstan', "Lao People's Democratic Republic", 'Latvia', 'Lebanon', 'Lesotho', 'Liberia', 'Libya', 'Liechtenstein', 'Lithuania', 'Luxembourg', 'Macao', 'Macedonia, the former Yugoslav Republic of', 'Madagascar', 'Malawi', 'Malaysia', 'Maldives', 'Mali', 'Malta', 'Marshall Islands', 'Martinique', 'Mauritania', 'Mauritius', 'Mayotte', 'Mexico', 'Micronesia, Federated States of', 'Moldova, Republic of', 'Monaco', 'Mongolia', 'Montenegro', 'Montserrat', 'Morocco', 'Mozambique', 'Myanmar', 'Namibia', 'Nauru', 'Nepal', 'Netherlands', 'New Caledonia', 'New Zealand', 'Nicaragua', 'Niger', 'Nigeria', 'Niue', 'Norfolk Island', 'Northern Mariana Islands', 'Norway', 'Oman', 'Pakistan', 'Palau', 'Palestine, State of', 'Panama', 'Papua New Guinea', 'Paraguay', 'Peru', 'Philippines', 'Pitcairn', 'Poland', 'Portugal', 'Puerto Rico', 'Qatar', 'Reunion', 'Romania', 'Russian Federation', 'Rwanda', 'Saint Barthelemy', 'Saint Helena, Ascension and Tristan da Cunha', 'Saint Kitts and Nevis', 'Saint Lucia', 'Saint Martin (French part)', 'Saint Pierre and Miquelon', 'Saint Vincent and the Grenadines', 'Samoa', 'San Marino', 'Sao Tome and Principe', 'Saudi Arabia', 'Senegal', 'Serbia', 'Seychelles', 'Sierra Leone', 'Singapore', 'Sint Maarten (Dutch part)', 'Slovakia', 'Slovenia', 'Solomon Islands', 'Somalia', 'South Africa', 'South Georgia and the South Sandwich Islands', 'South Sudan', 'Spain', 'Sri Lanka', 'Sudan', 'Suriname', 'Svalbard and Jan Mayen', 'Swaziland', 'Sweden', 'Switzerland', 'Syrian Arab Republic', 'Taiwan, Province of China', 'Tajikistan', 'Tanzania, United Republic of', 'Thailand', 'Timor-Leste', 'Togo', 'Tokelau', 'Tonga', 'Trinidad and Tobago', 'Tunisia', 'Turkey', 'Turkmenistan', 'Turks and Caicos Islands', 'Tuvalu', 'Uganda', 'Ukraine', 'United Arab Emirates', 'United Kingdom', 'United States', 'United States Minor Outlying Islands', 'Uruguay', 'Uzbekistan', 'Vanuatu', 'Venezuela, Bolivarian Republic of', 'Viet Nam', 'Virgin Islands, British', 'Virgin Islands, U.S.', 'Wallis and Futuna', 'Western Sahara', 'Yemen', 'Zambia', 'Zimbabwe', '')]
    [string] $Country,
    
    # Timezone for the new location. Acceptable values are based on the ISO list of timezones.
    [Parameter(ValueFromPipelineByPropertyName)]
    [AllowNull()]
    [AllowEmptyString()]
    [ValidateLength(0, 32)]
    [ValidateSet('ACT', 'AET', 'AGT', 'ART', 'AST', 'Africa/Abidjan', 'Africa/Accra', 'Africa/Addis_Ababa', 'Africa/Algiers', 'Africa/Asmara', 'Africa/Asmera', 'Africa/Bamako', 'Africa/Bangui', 'Africa/Banjul', 'Africa/Bissau', 'Africa/Blantyre', 'Africa/Brazzaville', 'Africa/Bujumbura', 'Africa/Cairo', 'Africa/Casablanca', 'Africa/Ceuta', 'Africa/Conakry', 'Africa/Dakar', 'Africa/Dar_es_Salaam', 'Africa/Djibouti', 'Africa/Douala', 'Africa/El_Aaiun', 'Africa/Freetown', 'Africa/Gaborone', 'Africa/Harare', 'Africa/Johannesburg', 'Africa/Juba', 'Africa/Kampala', 'Africa/Khartoum', 'Africa/Kigali', 'Africa/Kinshasa', 'Africa/Lagos', 'Africa/Libreville', 'Africa/Lome', 'Africa/Luanda', 'Africa/Lubumbashi', 'Africa/Lusaka', 'Africa/Malabo', 'Africa/Maputo', 'Africa/Maseru', 'Africa/Mbabane', 'Africa/Mogadishu', 'Africa/Monrovia', 'Africa/Nairobi', 'Africa/Ndjamena', 'Africa/Niamey', 'Africa/Nouakchott', 'Africa/Ouagadougou', 'Africa/Porto-Novo', 'Africa/Sao_Tome', 'Africa/Timbuktu', 'Africa/Tripoli', 'Africa/Tunis', 'Africa/Windhoek', 'America/Adak', 'America/Anchorage', 'America/Anguilla', 'America/Antigua', 'America/Araguaina', 'America/Argentina/Buenos_Aires', 'America/Argentina/Catamarca', 'America/Argentina/ComodRivadavia', 'America/Argentina/Cordoba', 'America/Argentina/Jujuy', 'America/Argentina/La_Rioja', 'America/Argentina/Mendoza', 'America/Argentina/Rio_Gallegos', 'America/Argentina/Salta', 'America/Argentina/San_Juan', 'America/Argentina/San_Luis', 'America/Argentina/Tucuman', 'America/Argentina/Ushuaia', 'America/Aruba', 'America/Asuncion', 'America/Atikokan', 'America/Atka', 'America/Bahia', 'America/Bahia_Banderas', 'America/Barbados', 'America/Belem', 'America/Belize', 'America/Blanc-Sablon', 'America/Boa_Vista', 'America/Bogota', 'America/Boise', 'America/Buenos_Aires', 'America/Cambridge_Bay', 'America/Campo_Grande', 'America/Cancun', 'America/Caracas', 'America/Catamarca', 'America/Cayenne', 'America/Cayman', 'America/Chicago', 'America/Chihuahua', 'America/Coral_Harbour', 'America/Cordoba', 'America/Costa_Rica', 'America/Creston', 'America/Cuiaba', 'America/Curacao', 'America/Danmarkshavn', 'America/Dawson', 'America/Dawson_Creek', 'America/Denver', 'America/Detroit', 'America/Dominica', 'America/Edmonton', 'America/Eirunepe', 'America/El_Salvador', 'America/Ensenada', 'America/Fort_Wayne', 'America/Fortaleza', 'America/Glace_Bay', 'America/Godthab', 'America/Goose_Bay', 'America/Grand_Turk', 'America/Grenada', 'America/Guadeloupe', 'America/Guatemala', 'America/Guayaquil', 'America/Guyana', 'America/Halifax', 'America/Havana', 'America/Hermosillo', 'America/Indiana/Indianapolis', 'America/Indiana/Knox', 'America/Indiana/Marengo', 'America/Indiana/Petersburg', 'America/Indiana/Tell_City', 'America/Indiana/Vevay', 'America/Indiana/Vincennes', 'America/Indiana/Winamac', 'America/Indianapolis', 'America/Inuvik', 'America/Iqaluit', 'America/Jamaica', 'America/Jujuy', 'America/Juneau', 'America/Kentucky/Louisville', 'America/Kentucky/Monticello', 'America/Knox_IN', 'America/Kralendijk', 'America/La_Paz', 'America/Lima', 'America/Los_Angeles', 'America/Louisville', 'America/Lower_Princes', 'America/Maceio', 'America/Managua', 'America/Manaus', 'America/Marigot', 'America/Martinique', 'America/Matamoros', 'America/Mazatlan', 'America/Mendoza', 'America/Menominee', 'America/Merida', 'America/Metlakatla', 'America/Mexico_City', 'America/Miquelon', 'America/Moncton', 'America/Monterrey', 'America/Montevideo', 'America/Montreal', 'America/Montserrat', 'America/Nassau', 'America/New_York', 'America/Nipigon', 'America/Nome', 'America/Noronha', 'America/North_Dakota/Beulah', 'America/North_Dakota/Center', 'America/North_Dakota/New_Salem', 'America/Ojinaga', 'America/Panama', 'America/Pangnirtung', 'America/Paramaribo', 'America/Phoenix', 'America/Port-au-Prince', 'America/Port_of_Spain', 'America/Porto_Acre', 'America/Porto_Velho', 'America/Puerto_Rico', 'America/Rainy_River', 'America/Rankin_Inlet', 'America/Recife', 'America/Regina', 'America/Resolute', 'America/Rio_Branco', 'America/Rosario', 'America/Santa_Isabel', 'America/Santarem', 'America/Santiago', 'America/Santo_Domingo', 'America/Sao_Paulo', 'America/Scoresbysund', 'America/Shiprock', 'America/Sitka', 'America/St_Barthelemy', 'America/St_Johns', 'America/St_Kitts', 'America/St_Lucia', 'America/St_Thomas', 'America/St_Vincent', 'America/Swift_Current', 'America/Tegucigalpa', 'America/Thule', 'America/Thunder_Bay', 'America/Tijuana', 'America/Toronto', 'America/Tortola', 'America/Vancouver', 'America/Virgin', 'America/Whitehorse', 'America/Winnipeg', 'America/Yakutat', 'America/Yellowknife', 'Antarctica/Casey', 'Antarctica/Davis', 'Antarctica/DumontDUrville', 'Antarctica/Macquarie', 'Antarctica/Mawson', 'Antarctica/McMurdo', 'Antarctica/Palmer', 'Antarctica/Rothera', 'Antarctica/South_Pole', 'Antarctica/Syowa', 'Antarctica/Vostok', 'Arctic/Longyearbyen', 'Asia/Aden', 'Asia/Almaty', 'Asia/Amman', 'Asia/Anadyr', 'Asia/Aqtau', 'Asia/Aqtobe', 'Asia/Ashgabat', 'Asia/Ashkhabad', 'Asia/Baghdad', 'Asia/Bahrain', 'Asia/Baku', 'Asia/Bangkok', 'Asia/Beirut', 'Asia/Bishkek', 'Asia/Brunei', 'Asia/Calcutta', 'Asia/Choibalsan', 'Asia/Chongqing', 'Asia/Chungking', 'Asia/Colombo', 'Asia/Dacca', 'Asia/Damascus', 'Asia/Dhaka', 'Asia/Dili', 'Asia/Dubai', 'Asia/Dushanbe', 'Asia/Gaza', 'Asia/Harbin', 'Asia/Hebron', 'Asia/Ho_Chi_Minh', 'Asia/Hong_Kong', 'Asia/Hovd', 'Asia/Irkutsk', 'Asia/Istanbul', 'Asia/Jakarta', 'Asia/Jayapura', 'Asia/Jerusalem', 'Asia/Kabul', 'Asia/Kamchatka', 'Asia/Karachi', 'Asia/Kashgar', 'Asia/Kathmandu', 'Asia/Katmandu', 'Asia/Kolkata', 'Asia/Krasnoyarsk', 'Asia/Kuala_Lumpur', 'Asia/Kuching', 'Asia/Kuwait', 'Asia/Macao', 'Asia/Macau', 'Asia/Magadan', 'Asia/Makassar', 'Asia/Manila', 'Asia/Muscat', 'Asia/Nicosia', 'Asia/Novokuznetsk', 'Asia/Novosibirsk', 'Asia/Omsk', 'Asia/Oral', 'Asia/Phnom_Penh', 'Asia/Pontianak', 'Asia/Pyongyang', 'Asia/Qatar', 'Asia/Qyzylorda', 'Asia/Rangoon', 'Asia/Riyadh', 'Asia/Riyadh87', 'Asia/Riyadh88', 'Asia/Riyadh89', 'Asia/Saigon', 'Asia/Sakhalin', 'Asia/Samarkand', 'Asia/Seoul', 'Asia/Shanghai', 'Asia/Singapore', 'Asia/Taipei', 'Asia/Tashkent', 'Asia/Tbilisi', 'Asia/Tehran', 'Asia/Tel_Aviv', 'Asia/Thimbu', 'Asia/Thimphu', 'Asia/Tokyo', 'Asia/Ujung_Pandang', 'Asia/Ulaanbaatar', 'Asia/Ulan_Bator', 'Asia/Urumqi', 'Asia/Vientiane', 'Asia/Vladivostok', 'Asia/Yakutsk', 'Asia/Yekaterinburg', 'Asia/Yerevan', 'Atlantic/Azores', 'Atlantic/Bermuda', 'Atlantic/Canary', 'Atlantic/Cape_Verde', 'Atlantic/Faeroe', 'Atlantic/Faroe', 'Atlantic/Jan_Mayen', 'Atlantic/Madeira', 'Atlantic/Reykjavik', 'Atlantic/South_Georgia', 'Atlantic/St_Helena', 'Atlantic/Stanley', 'Australia/ACT', 'Australia/Adelaide', 'Australia/Brisbane', 'Australia/Broken_Hill', 'Australia/Canberra', 'Australia/Currie', 'Australia/Darwin', 'Australia/Eucla', 'Australia/Hobart', 'Australia/LHI', 'Australia/Lindeman', 'Australia/Lord_Howe', 'Australia/Melbourne', 'Australia/NSW', 'Australia/North', 'Australia/Perth', 'Australia/Queensland', 'Australia/South', 'Australia/Sydney', 'Australia/Tasmania', 'Australia/Victoria', 'Australia/West', 'Australia/Yancowinna', 'BET', 'BST', 'Brazil/Acre', 'Brazil/DeNoronha', 'Brazil/East', 'Brazil/West', 'CAT', 'CET', 'CNT', 'CST', 'CST6CDT', 'CTT', 'Canada/Atlantic', 'Canada/Central', 'Canada/East-Saskatchewan', 'Canada/Eastern', 'Canada/Mountain', 'Canada/Newfoundland', 'Canada/Pacific', 'Canada/Saskatchewan', 'Canada/Yukon', 'Chile/Continental', 'Chile/EasterIsland', 'Cuba', 'EAT', 'ECT', 'EET', 'EST', 'EST5EDT', 'Egypt', 'Eire', 'Etc/GMT', 'Etc/GMT+0', 'Etc/GMT+1', 'Etc/GMT+10', 'Etc/GMT+11', 'Etc/GMT+12', 'Etc/GMT+2', 'Etc/GMT+3', 'Etc/GMT+4', 'Etc/GMT+5', 'Etc/GMT+6', 'Etc/GMT+7', 'Etc/GMT+8', 'Etc/GMT+9', 'Etc/GMT-0', 'Etc/GMT-1', 'Etc/GMT-10', 'Etc/GMT-11', 'Etc/GMT-12', 'Etc/GMT-13', 'Etc/GMT-14', 'Etc/GMT-2', 'Etc/GMT-3', 'Etc/GMT-4', 'Etc/GMT-5', 'Etc/GMT-6', 'Etc/GMT-7', 'Etc/GMT-8', 'Etc/GMT-9', 'Etc/GMT0', 'Etc/Greenwich', 'Etc/UCT', 'Etc/UTC', 'Etc/Universal', 'Etc/Zulu', 'Europe/Amsterdam', 'Europe/Andorra', 'Europe/Athens', 'Europe/Belfast', 'Europe/Belgrade', 'Europe/Berlin', 'Europe/Bratislava', 'Europe/Brussels', 'Europe/Bucharest', 'Europe/Budapest', 'Europe/Chisinau', 'Europe/Copenhagen', 'Europe/Dublin', 'Europe/Gibraltar', 'Europe/Guernsey', 'Europe/Helsinki', 'Europe/Isle_of_Man', 'Europe/Istanbul', 'Europe/Jersey', 'Europe/Kaliningrad', 'Europe/Kiev', 'Europe/Lisbon', 'Europe/Ljubljana', 'Europe/London', 'Europe/Luxembourg', 'Europe/Madrid', 'Europe/Malta', 'Europe/Mariehamn', 'Europe/Minsk', 'Europe/Monaco', 'Europe/Moscow', 'Europe/Nicosia', 'Europe/Oslo', 'Europe/Paris', 'Europe/Podgorica', 'Europe/Prague', 'Europe/Riga', 'Europe/Rome', 'Europe/Samara', 'Europe/San_Marino', 'Europe/Sarajevo', 'Europe/Simferopol', 'Europe/Skopje', 'Europe/Sofia', 'Europe/Stockholm', 'Europe/Tallinn', 'Europe/Tirane', 'Europe/Tiraspol', 'Europe/Uzhgorod', 'Europe/Vaduz', 'Europe/Vatican', 'Europe/Vienna', 'Europe/Vilnius', 'Europe/Volgograd', 'Europe/Warsaw', 'Europe/Zagreb', 'Europe/Zaporozhye', 'Europe/Zurich', 'Factory', 'GB', 'GB-Eire', 'GMT', 'GMT+0', 'GMT-0', 'GMT0', 'Greenwich', 'HST', 'Hongkong', 'IET', 'IST', 'Iceland', 'Indian/Antananarivo', 'Indian/Chagos', 'Indian/Christmas', 'Indian/Cocos', 'Indian/Comoro', 'Indian/Kerguelen', 'Indian/Mahe', 'Indian/Maldives', 'Indian/Mauritius', 'Indian/Mayotte', 'Indian/Reunion', 'Iran', 'Israel', 'JST', 'Jamaica', 'Japan', 'Kwajalein', 'Libya', 'MET', 'MIT', 'MST', 'MST7MDT', 'Mexico/BajaNorte', 'Mexico/BajaSur', 'Mexico/General', 'Mideast/Riyadh87', 'Mideast/Riyadh88', 'Mideast/Riyadh89', 'NET', 'NST', 'NZ', 'NZ-HAT', 'Navajo', 'PLT', 'PNT', 'PRC', 'PRT', 'PST', 'PST8PDT', 'Pacific/Apia', 'Pacific/Auckland', 'Pacific/Chatham', 'Pacific/Chuuk', 'Pacific/Easter', 'Pacific/Efate', 'Pacific/Enderbury', 'Pacific/Fakaofo', 'Pacific/Fiji', 'Pacific/Funafuti', 'Pacific/Galapagos', 'Pacific/Gambier', 'Pacific/Guadalcanal', 'Pacific/Guam', 'Pacific/Honolulu', 'Pacific/Johnston', 'Pacific/Kiritimati', 'Pacific/Kosrae', 'Pacific/Kwajalein', 'Pacific/Majuro', 'Pacific/Marquesas', 'Pacific/Midway', 'Pacific/Nauru', 'Pacific/Niue', 'Pacific/Norfolk', 'Pacific/Noumea', 'Pacific/Pago_Pago', 'Pacific/Palau', 'Pacific/Pitcairn', 'Pacific/Pohnpei', 'Pacific/Ponape', 'Pacific/Port_Moresby', 'Pacific/Rarotonga', 'Pacific/Saipan', 'Pacific/Samoa', 'Pacific/Tahiti', 'Pacific/Tarawa', 'Pacific/Tongatapu', 'Pacific/Truk', 'Pacific/Wake', 'Pacific/Wallis', 'Pacific/Yap', 'Poland', 'Portugal', 'ROC', 'ROK', 'SST', 'Singapore', 'SystemV/AST4', 'SystemV/AST4ADT', 'SystemV/CST6', 'SystemV/CST6CDT', 'SystemV/EST5', 'SystemV/EST5EDT', 'SystemV/HST10', 'SystemV/MST7', 'SystemV/MST7MDT', 'SystemV/PST8', 'SystemV/PST8PDT', 'SystemV/YST9', 'SystemV/YST9YDT', 'Turkey', 'UCT', 'US/Alaska', 'US/Aleutian', 'US/Arizona', 'US/Central', 'US/East-Indiana', 'US/Eastern', 'US/Hawaii', 'US/Indiana-Starke', 'US/Michigan', 'US/Mountain', 'US/Pacific', 'US/Pacific-New', 'US/Samoa', 'UTC', 'Universal', 'VST', 'W-SU', 'WET', 'Zulu', '')]
    [string] $TimeZone
  )
  Begin{
    ConnectToServerIfNeeded
  }
  Process{
    try{
      # Assemble the query string based on the parameters specified
      $Params = @{"Identity" = $Identity}
      if ($Name){$Params["Name"] = $Name}
      if ($Note){$Params["Note"] = $Note}
      if ($ContactName){$Params["ContactName"] = $ContactName}
      if ($ContactNumber){$Params["ContactNumber"] = $ContactNumber}
      if ($ContactEmail){$Params["ContactEmail"] = $ContactEmail}
      if ($Address){$Params["Address"] = $Address}
      if ($City){$Params["City"] = $City}
      if ($State){$Params["State"] = $State}
      if ($PostalCode){$Params["PostalCode"] = $PostalCode}
      if ($Country){$Params["Country"] = $Country}
      if ($TimeZone){$Params["TimeZone"] = $TimeZone}
      
      # What-If suport
      if (-not $(ShouldProcess $Params)) {
        return
      }

      # Assemble the URI
      [string] $Uri = $HTTP + $global:EzUdmTenant + '/' + $EzUdmLocationApi
      Write-Verbose -Message $Uri
      $WebRequest = Invoke-WebRequest -Uri $Uri -Method Post -Credential $global:EzUdmCreds -Body $Params
    
      # Handle the JSON data returned
      # This code is just working theory right now, and should not be trusted
      $parsed = ConvertFrom-Json -InputObject $WebRequest.Content 
      return $parsed
    }
    catch{
      Throw "An error has occured. The error was: '$_'."
    }
  }
  End{}
} # end function Set-UdmLocation
function Get-UdmLocation{
  <#
      .SYNOPSIS
      Retrieves a single or all Locations within the core UDM product.

      .DESCRIPTION
      Retrieves a single or all Locations within the core UDM product.

      .NOTES
      Version 1.0

      .INPUTS
      List of input types that are accepted by this function.

      .OUTPUTS
      system.object
  #>


  [CmdletBinding(SupportsShouldProcess, SupportsPaging, HelpURI = '')]
  param(
    # Identity of the Location to be retrieved. If no Location is specified, all Locations will be returned.
    [Parameter(ValueFromPipeline, ValueFromPipelineByPropertyName)]
    [ValidateLength(1,40)]
    [string] $Identity
  )
  Begin{
    ConnectToServerIfNeeded
  }
  Process{
    try{
      # Assemble the query string based on the parameters specified
      if ($Identity){$Query = "Identity=$Identity&"}

      # What-If suport
      if (-not $(ShouldProcess $Query)) {
        return
      } 

      # Assemble the URI
      [string] $Uri = $HTTP + $global:EzUdmTenant + '/' + $EzUdmLocationApi + '?' + $Query
      Write-Verbose -Message $Uri
      $WebRequest = Invoke-WebRequest -Uri $Uri -Method Get -Credential $global:EzUdmCreds
    
      # Handle the JSON data returned
      # This code is just working theory right now, and should not be trusted
      $parsed = ConvertFrom-Json -InputObject $WebRequest.Content 
      return $parsed
    }
    catch{
      Throw "An error has occured. The error was: '$_'."
    }
  }
  End{}
} # end function Get-UdmLocation
function Remove-UdmLocation{
  <#
      .SYNOPSIS
      Removes the specified Location from the core UDM product.

      .DESCRIPTION
      Removes the specified Location from the core UDM product.

      .NOTES
      Version 1.0

      .INPUTS
      List of input types that are accepted by this function.

      .OUTPUTS
      system.object
  #>


  [CmdletBinding(SupportsShouldProcess, SupportsPaging, HelpURI = '')]
  param(
    # Identity of the Location to be deleted.
    [Parameter(Mandatory=$true, ValueFromPipeline, ValueFromPipelineByPropertyName)]
    [Alias('id')]
    [ValidateNotNullOrEmpty()]
    [ValidateLength(1,40)]
    [string] $Identity
  )
  Begin{
    ConnectToServerIfNeeded
  }
  Process{
    try{
      # Assemble the query string based on the parameters specified
      if ($Identity){$Query = "Identity=$Identity"}

      # What-If suport
      if (-not $(ShouldProcess $Query)) {
        return
      }

      # Assemble the URI
      [string] $Uri = $HTTP + $global:EzUdmTenant + '/' + $EzUdmLocationApi + '?' + $Query
      Write-Verbose -Message $Uri
      $WebRequest = Invoke-WebRequest -Uri $Uri -Method Delete -Credential $global:EzUdmCreds
    
      return $WebRequest.Content
    }
    catch{
      Throw "An error has occured. The error was: '$_'."
    }
  }
  End{}
} # end function Remove-UdmLocation
#endregion UdmLocation

#region UdmRmRange
function New-UdmRmRange{
  <#
      .SYNOPSIS
      Creates a new Range in the Range Manager product.

      .DESCRIPTION
      Creates a new Range in the Range Manager product.

      .EXAMPLE
      New-UdmRmRange -Name name -Notes note 
      
      .NOTES
      Version 1.0

      .INPUTS
      List of input types that are accepted by this function.

      .OUTPUTS
      system.object
  #>


  [CmdletBinding(SupportsShouldProcess, SupportsPaging, HelpURI = '')]
  param(
    # Name of the Range to be created
    [Parameter(Mandatory=$true, ValueFromPipeline)]
    [ValidateNotNullOrEmpty()]
    [ValidateLength(1,40)]
    [string] $Name,
    
    # Freeform notes to assign to this Range
    [Parameter(ValueFromPipelineByPropertyName)]
    [ValidateLength(1,255)]
    [string] $Note
  )
  Begin{
    ConnectToServerIfNeeded
  }
  Process{
    try{
      $Params = @{}
      if ($Name){$Params["Name"] = $Name}
      if ($Note){$Params["Note"] = $Note}

      # What-If suport
      if (-not $(ShouldProcess $Params)) {
        return
      }

      # Assemble the URI
      [string] $Uri = $HTTP + $global:EzUdmTenant + '/' + $EzUdmRmRangeApi
      Write-Verbose -Message $Uri
      $WebRequest = Invoke-WebRequest -Uri $Uri -Method Post -Credential $global:EzUdmCreds -Body $Params
    
      $parsed = ConvertFrom-Json -InputObject $WebRequest.Content 
      return $parsed
    }
    catch{
      Throw "An error has occured. The error was: '$_'."
    }
  }
  End{}
} # end function New-UdmRmRange
function Set-UdmRmRange{
  <#
      .SYNOPSIS
      Updates an existing Range within the Range Manager product.

      .DESCRIPTION
      Updates an existing Range within the Range Manager product.

      .EXAMPLE
      Set-UdmRmRange -Identity uuid -Notes note 
      

      .NOTES
      Version 1.0

      .INPUTS
      List of input types that are accepted by this function.

      .OUTPUTS
      system.object
  #>


  [CmdletBinding(SupportsShouldProcess, SupportsPaging, HelpURI = '')]
  param(
    # Identity of the Range to be updated
    [Parameter(Mandatory=$true, ValueFromPipelineByPropertyName)]
    [Alias('id')]
    [ValidateNotNullOrEmpty()]
    [ValidateLength(1,40)]
    [string] $Identity,

    # Name of the Range to be created
    [Parameter(ValueFromPipelineByPropertyName)]
    [ValidateNotNullOrEmpty()]
    [ValidateLength(1,40)]
    [string] $Name,

    # Freeform notes to assign to this 
    [Parameter(ValueFromPipelineByPropertyName)]
    [ValidateLength(1,255)]
    [string] $Note
  )
  Begin{
    ConnectToServerIfNeeded
  }
  Process{
    try{
      # Assemble the query string based on the parameters specified
      $Params = @{"Identity" = $Identity}
      if ($Name){$Params["Name"] = $Name}
      if ($Note){$Params["Note"] = $Note}
      
      # What-If suport
      if (-not $(ShouldProcess $Params)) {
        return
      }

      # Assemble the URI
      [string] $Uri = $HTTP + $global:EzUdmTenant + '/' + $EzUdmRmRangeApi
      Write-Verbose -Message $Uri
      $WebRequest = Invoke-WebRequest -Uri $Uri -Method Post -Credential $global:EzUdmCreds -Body $Params
    
      # Handle the JSON data returned
      # This code is just working theory right now, and should not be trusted
      $parsed = ConvertFrom-Json -InputObject $WebRequest.Content 
      return $parsed
    }
    catch{
      Throw "An error has occured. The error was: '$_'."
    }
  }
  End{}
} # end function Set-UdmRmRange
function Get-UdmRmRange{
  <#
      .SYNOPSIS
      Retrieves a single, or all, Ranges within the Range Manager product.

      .DESCRIPTION
      Retrieves a single, or all, Ranges within the Range Manager product.

      .EXAMPLE
      Get-UdmRmRange -Identity Value
      Describe what this call does

      .NOTES
      Version 1.0

      .INPUTS
      List of input types that are accepted by this function.

      .OUTPUTS
      system.object
  #>


  [CmdletBinding(SupportsShouldProcess, SupportsPaging, HelpURI = '')]
  param(
    # Identity of the Range to be retrieved. If no Identity is specified, all Ranges will be returned.
    [Parameter(ValueFromPipeline, ValueFromPipelineByPropertyName)]
    [ValidateLength(1,40)]
    [string] $Identity
  )
  Begin{
    ConnectToServerIfNeeded
  }
  Process{
    try{
      # Assemble the query string based on the parameters specified
      if ($Identity){$Query = "Identity=$Identity&"}

      # What-If suport
      if (-not $(ShouldProcess $Query)) {
        return
      }

      # Assemble the URI
      [string] $Uri = $HTTP + $global:EzUdmTenant + '/' + $EzUdmRmRangeApi + '?' + $Query
      Write-Verbose -Message $Uri
      $WebRequest = Invoke-WebRequest -Uri $Uri -Method Get -Credential $global:EzUdmCreds
    
      # Handle the JSON data returned
      # This code is just working theory right now, and should not be trusted
      $parsed = ConvertFrom-Json -InputObject $WebRequest.Content 
      return $parsed
    }
    catch{
      Throw "An error has occured. The error was: '$_'."
    }
  }
  End{}
} # end function Get-UdmRmRange
function Remove-UdmRmRange{
  <#
      .SYNOPSIS
      Removes an existing Range within the Range Manager product.

      .DESCRIPTION
      Removes an existing Range within the Range Manager product.

      .EXAMPLE
      Remove-UdmRmRange -Identity Value
      Describe what this call does

      .NOTES
      Version 1.0

      .INPUTS
      List of input types that are accepted by this function.

      .OUTPUTS
      system.object
  #>


  [CmdletBinding(SupportsShouldProcess, SupportsPaging, HelpURI = '')]
  param(
    # Identity of the Range to be removed
    [Parameter(Mandatory=$true,ValueFromPipeline, ValueFromPipelineByPropertyName)]
    [Alias('id')]
    [ValidateNotNullOrEmpty()]
    [ValidateLength(1,40)]
    [string] $Identity
  )
  Begin{
    ConnectToServerIfNeeded
  }
  Process{
    try{
      # Assemble the query string based on the parameters specified
      if ($Identity){$Query = "Identity=$Identity"}

      # What-If suport
      if (-not $(ShouldProcess $Query)) {
        return
      }

      # Assemble the URI
      [string] $Uri = $HTTP + $global:EzUdmTenant + '/' + $EzUdmRmRangeApi + '?' + $Query
      Write-Verbose -Message $Uri
      $WebRequest = Invoke-WebRequest -Uri $Uri -Method Delete -Credential $global:EzUdmCreds
    
      return $WebRequest.Content
    }
    catch{
      Throw "An error has occured. The error was: '$_'."
    }
  }
  End{}
} # end function Remove-UdmRmRange
#endregion UdmRmRange

#region UdmRmNumberBlock
function New-UdmRmNumberBlock{
  <#
      .SYNOPSIS
      Creates a new Number Block within an existing Range in the Ranger Manager product.

      .DESCRIPTION
      Creates a new Number Block within an existing Range in the Ranger Manager product.

      .EXAMPLE
      New-UdmRmNumberBlock -Name Value -Notes Value
      Describe what this call does

      .NOTES
      Version 1.0

      .INPUTS
      List of input types that are accepted by this function.

      .OUTPUTS
      system.object
  #>


  [CmdletBinding(SupportsShouldProcess, SupportsPaging, HelpURI = '')]
  param(
    # Range that the new Number Block is for
    [Parameter(Mandatory=$true, ValueFromPipelineByPropertyName)]
    [ValidateNotNullOrEmpty()]
    [ValidateLength(1,40)]
    [string] $RangeIdentity,

    [Parameter(Mandatory=$true, ValueFromPipelineByPropertyName)]
    [ValidateNotNullOrEmpty()]
    [ValidateScript({If ($_ -imatch $PHONE_NUMBER_FORMAT) {
      $True
    } Else {
      Throw "'$_' is not a valid number"
    }})]
    [string] $PhoneNumber,

    [Parameter(ValueFromPipelineByPropertyName)]
    [ValidateScript({If ($_ -imatch $EXTENSION_FORMAT) {
      $True
    } Else {
      Throw "'$_' is not a valid extension"
    }})]
    [string] $Extension,

    [Parameter(ValueFromPipelineByPropertyName)]
    [ValidateScript({If ($_ -imatch $QUANTITY_FORMAT) {
      $True
    } Else {
      Throw "'$_'' must be between 1 and 999999"
    }})]
    [string] $Quantity,

    # Service ID to be assigned to the new Range
    [Parameter(ValueFromPipelineByPropertyName)]
    [ValidateLength(1,40)]
    [string] $ServiceId,

    # Service Provider to be assigned to the new Range
    [Parameter(ValueFromPipelineByPropertyName)]
    [ValidateLength(1,40)]
    [string] $ServiceProvider,

    # Freeform notes to assign to this 
    [Parameter(ValueFromPipelineByPropertyName)]
    [ValidateLength(1,255)]
    [string] $Note
  )
  Begin{
    ConnectToServerIfNeeded
  }
  Process{
    try{
      $Params = @{}
      if ($RangeIdentity){$Params["RangeIdentity"] = $RangeIdentity}
      if ($PhoneNumber){$Params["PhoneNumber"] = $PhoneNumber}
      if ($Extension){$Params["Extension"] = $Extension}
      if ($Quantity){$Params["Quantity"] = $Quantity}
      if ($ServiceId){$Params["ServiceId"] = $ServiceId}
      if ($ServiceProvider){$Params["ServiceProvider"] = $ServiceProvider}
      if ($Note){$Params["Note"] = $Note}

      # What-If suport
      if (-not $(ShouldProcess $Params)) {
        return
      }

      # Assemble the URI
      [string] $Uri = $HTTP + $global:EzUdmTenant + '/' + $EzUdmRmNumberBlockApi
      Write-Host $Uri
      Write-Verbose -Message $Uri
      $WebRequest = Invoke-WebRequest -Uri $Uri -Method Post -Credential $global:EzUdmCreds -Body $Params
    
      $parsed = ConvertFrom-Json -InputObject $WebRequest.Content 
      return $parsed
    }
    catch{
      Throw "An error has occured. The error was: '$_'."
    }
  }
  End{}
} # end function New-UdmRmNumberBlock
function Set-UdmRmNumberBlock{
  <#
      .SYNOPSIS
      Updates an existing Number Block within the Range Manager product.

      .DESCRIPTION
      Updates an existing Number Block within the Range Manager product.

      .EXAMPLE
      Set-UdmRmNumberBlock -Identity Value -Notes Value
      Describe what this call does

      .NOTES
      Version 1.0

      .INPUTS
      List of input types that are accepted by this function.

      .OUTPUTS
      system.object
  #>


  [CmdletBinding(SupportsShouldProcess, SupportsPaging, HelpURI = '')]
  param(
    # Identity of the Number Block to be updated.
    [Parameter(Mandatory=$true, ValueFromPipelineByPropertyName)]
    [Alias('id')]
    [ValidateNotNullOrEmpty()]
    [ValidateLength(1,40)]
    [string] $Identity,

    # Range that the Number Block is for, which is mandatory
    [Parameter(Mandatory=$true, ValueFromPipelineByPropertyName)]
    [Alias('rangeId')]
    [ValidateLength(1,40)]
    [string] $RangeIdentity,

    [Parameter(Mandatory=$true, ValueFromPipelineByPropertyName)]
    [ValidateScript({If ($_ -imatch $PHONE_NUMBER_FORMAT) {
      $True
    } Else {
      Throw "$_ is not a valid number"
    }})]
    [string] $PhoneNumber,

    [Parameter(Mandatory=$true, ValueFromPipelineByPropertyName)]
    [AllowEmptyString()]
    [ValidateScript({If ($_ -imatch $EXTENSION_FORMAT) {
      $True
    } Else {
      Throw "$_ is not a valid extension"
    }})]
    [string] $Extension,

    [Parameter(Mandatory=$true, ValueFromPipelineByPropertyName)]
    [ValidateScript({If ($_ -imatch $QUANTITY_FORMAT) {
      $True
    } Else {
      Throw "$_ must be between 1 and 999999"
    }})]
    [string] $Quantity,

    # Service ID to be assigned to the new Range
    [Parameter(ValueFromPipelineByPropertyName)]
    [ValidateLength(0,40)]
    [string] $ServiceId,

    # Service Provider to be assigned to the new Range
    [Parameter(ValueFromPipelineByPropertyName)]
    [ValidateLength(0,40)]
    [string] $ServiceProvider,
    
    # New freeform notes to assign to this Number Block
    [Parameter(ValueFromPipelineByPropertyName)]
    [ValidateLength(0,255)]
    [string] $Note
  )
  Begin{
    ConnectToServerIfNeeded
  }
  Process{
    try{
      $Params = @{}
      $Params["Identity"] = $Identity
      if ($RangeIdentity){$Params["RangeIdentity"] = $RangeIdentity}
      if ($PhoneNumber){$Params["PhoneNumber"] = $PhoneNumber}
      if ($Extension){$Params["Extension"] = $Extension}
      if ($Quantity){$Params["Quantity"] = $Quantity}
      if ($ServiceId){$Params["ServiceId"] = $ServiceId}
      if ($ServiceProvider){$Params["ServiceProvider"] = $ServiceProvider}
      if ($Note){$Params["Note"] = $Note}

      # What-If suport
      if (-not $(ShouldProcess $Params)) {
        return
      }

      # Assemble the URI
      [string] $Uri = $HTTP + $global:EzUdmTenant + '/' + $EzUdmRmNumberBlockApi
      Write-Verbose -Message $Uri
      $WebRequest = Invoke-WebRequest -Uri $Uri -Method Post -Credential $global:EzUdmCreds -Body $Params
    
      $parsed = ConvertFrom-Json -InputObject $WebRequest.Content 
      return $parsed
    }
    catch{
      Throw "An error has occured. The error was: '$_'."
    }
  }
  End{}
} # end function Set-UdmRmNumberBlock
function Get-UdmRmNumberBlock{
  <#
      .SYNOPSIS
      Retrieves a single, or all, Number Blocks within the Range Manager product.

      .DESCRIPTION
      Retrieves a single, or all, Number Blocks within the Range Manager product.

      .EXAMPLE
      Get-UdmRmNumberBlock -Identity Value
      Describe what this call does

      .NOTES
      Version 1.0

      .INPUTS
      List of input types that are accepted by this function.

      .OUTPUTS
      system.object
  #>


  [CmdletBinding(SupportsShouldProcess, SupportsPaging, HelpURI = '')]
  param(
    # Identity of the Number Block to be retrieved. If no Number Block is specified, all Number Blocks will be retrieved.
    [Parameter(ValueFromPipeline, ValueFromPipelineByPropertyName)]
    [Alias('id')]
    [ValidateLength(1,40)]
    [string] $Identity
  )
  Begin{
    ConnectToServerIfNeeded
  }
  Process{
    try{
      # Assemble the query string based on the parameters specified
      if ($Identity){$Query = "Identity=$Identity&"}

      # What-If suport
      if (-not $(ShouldProcess $Query)) {
        return
      }   

      # Assemble the URI
      [string] $Uri = $HTTP + $global:EzUdmTenant + '/' + $EzUdmRmNumberBlockApi + '?' + $Query
      Write-Verbose -Message $Uri
      $WebRequest = Invoke-WebRequest -Uri $Uri -Method Get -Credential $global:EzUdmCreds
    
      $parsed = ConvertFrom-Json -InputObject $WebRequest.Content 
      return $parsed
    }
    catch{
      Throw "An error has occured. The error was: '$_'."
    }
  }
  End{}
} # end function Get-UdmRmNumberBlock
function Remove-UdmRmNumberBlock{
  <#
      .SYNOPSIS
      Removes an existing Number Block from the Range Manager product.

      .DESCRIPTION
      Removes an existing Number Block from the Range Manager product.

      .EXAMPLE
      Remove-UdmRmNumberBlock -Identity Value
      Describe what this call does

      .NOTES
      Version 1.0

      .INPUTS
      List of input types that are accepted by this function.

      .OUTPUTS
      system.object
  #>


  [CmdletBinding(SupportsShouldProcess, SupportsPaging, HelpURI = '')]
  param(
    # Identity of the Number Block to be deleted.
    [Parameter(Mandatory=$true, ValueFromPipeline, ValueFromPipelineByPropertyName)]
    [Alias('id')]
    [ValidateNotNullOrEmpty()]
    [ValidateLength(1,40)]
    [string] $Identity
  )
  Begin{
    ConnectToServerIfNeeded
  }
  Process{
    try{
      # Assemble the query string based on the parameters specified
      if ($Identity){$Query = "Identity=$Identity&"}

      # What-If suport
      if (-not $(ShouldProcess $Query)) {
        return
      }    

      # Assemble the URI
      [string] $Uri = $HTTP + $global:EzUdmTenant + '/' + $EzUdmRmNumberBlockApi + '?' + $Query
      Write-Verbose -Message $Uri
      $WebRequest = Invoke-WebRequest -Uri $Uri -Method Delete -Credential $global:EzUdmCreds

      return $WebRequest.Content
    }
    catch{
      Throw "An error has occured. The error was: '$_'."
    }
  }
  End{}
} # end function Remove-UdmRmNumberBlock
#endregion UdmRmNumberBlock

#region UdmRmNumber
function Get-UdmRmNumber{
  <#
      .SYNOPSIS
      Retrieves a single, or all, Numbers from the Range Manager product.

      .DESCRIPTION
      Retrieves a single, or all, Numbers from the Range Manager product.

      .EXAMPLE
      Get-UdmRmNumber -Identity phoneNumber 
      Describe what this call does

      .NOTES
      Version 1.0

      .INPUTS
      List of input types that are accepted by this function.

      .OUTPUTS
      system.object
  #>


  [CmdletBinding(SupportsShouldProcess, SupportsPaging, HelpURI = '')]
  param(
    # Identity of the Number to be retrieved. If no Number is specified, all Numbers will be retrieved.
    [Parameter(ValueFromPipeline, ValueFromPipelineByPropertyName)]
    [ValidateLength(1,40)]
    [string] $Identity
  )
  Begin{
    ConnectToServerIfNeeded
  }
  Process{
    try{
      # Assemble the query string based on the parameters specified
      if ($Identity){$Query = "Identity=$Identity&"}
    
      # What-If suport
      if (-not $(ShouldProcess $Query)) {
        return
      }

      # Assemble the URI
      [string] $Uri = $HTTP + $global:EzUdmTenant + '/' + $EzUdmRmNumberApi + '?' + $Query
      Write-Verbose -Message $Uri
      $WebRequest = Invoke-WebRequest -Uri $Uri -Method Get -Credential $global:EzUdmCreds
    
      # Handle the JSON data returned
      # This code is just working theory right now, and should not be trusted
      $parsed = ConvertFrom-Json -InputObject $WebRequest.Content
      return $parsed
    }
    catch{
      Throw "An error has occured. The error was: '$_'."
    }
  }
  End{}
} # end function Get-UdmRmNumber
function Set-UdmRmNumber{
  <#
      .SYNOPSIS
      Updates an existing Number within the Range Manager product.

      .DESCRIPTION
      Updates an existing Number within the Range Manager product.

      .EXAMPLE
      Set-UdmRmNumber -Identity Value -Notes Value -Reserved Value
      Describe what this call does

      .NOTES
      Version 1.0

      .INPUTS
      List of input types that are accepted by this function.

      .OUTPUTS
      system.object
  #>


  [CmdletBinding(SupportsShouldProcess, SupportsPaging, HelpURI = '')]
  param(
    # Identity of the Number to be updated.
    [Parameter(ValueFromPipelineByPropertyName)]
    [Alias('did')]
    [ValidateNotNullOrEmpty()]
    [ValidateLength(1,40)]
    [string] $Identity,
    
    # New freeform notes to assign to this Number
    [Parameter(ValueFromPipelineByPropertyName)]
    [ValidateLength(1,255)]
    [string] $Note,

    # New boolean for reserving a number
    [Parameter(ValueFromPipelineByPropertyName)]
    [bool] $Reserved
  )
  Begin{
    ConnectToServerIfNeeded
  }
  Process{
    try{
      # Assemble the query string based on the parameters specified
      if ($Identity){$Query = "Identity=$Identity&"}
      if ($Note){$Query += "Note=$Note&"}
      if ($Reserved -ne $null){$Query += "Reserved=$Reserved&"}

      # What-If suport
      if (-not $(ShouldProcess $Query)) {
        return
      }

      # Assemble the URI
      [string] $Uri = $HTTP + $global:EzUdmTenant + '/' + $EzUdmRmNumberApi + '?' + $Query
      Write-Verbose -Message $Uri
      $WebRequest = Invoke-WebRequest -Uri $Uri -Method Post -Credential $global:EzUdmCreds
    
      # Handle the JSON data returned
      # This code is just working theory right now, and should not be trusted
      $parsed = ConvertFrom-Json -InputObject $WebRequest.Content 
      # $parsed.origin
      # $parsed.headers
      return $parsed
    }
    catch{
      Throw "An error has occured. The error was: '$_'."
    }
  }
  End{}
} # end function Set-UdmRmNumber
#endregion UdmRmNumber

#region UdmRmNumberReservation
function New-UdmRmNumberReservation{
  <#
      .SYNOPSIS
      Reserves a Number within the Range Manager product.

      .DESCRIPTION
      Reserves a Number within the Range Manager product.

      .EXAMPLE
      New-UdmRmNumberReservation -Did
      Describe what this call does

      .EXAMPLE
      New-UdmRmNumberReservation -Ext
      Describe what this call does

      .EXAMPLE
      New-UdmRmNumberReservation -Range Value -ServiceProvider Value -ServiceId Value
      Describe what this call does

      .NOTES
      Version 1.0

      .INPUTS
      List of input types that are accepted by this function.

      .OUTPUTS
      system.object
  #>


  [CmdletBinding(SupportsShouldProcess, SupportsPaging, HelpURI = '', DefaultParameterSetName = 'Did')]
  param(
    # When specified, will reserve a number from that Range
    [Parameter(Mandatory=$true, ValueFromPipelineByPropertyName)]
    [ValidateLength(1,255)]
    [string] $Range,

    [Parameter(ValueFromPipelineByPropertyName)]
    [AllowEmptyString()]
    [ValidateScript({If ($_ -imatch $TIME_FORMAT) {
      $True
    } Else {
      Throw "$_ is not a valid reservation time. Enter a value from 0 to 9999999"
    }})]
    [string] $ReservedMinutes,

    # When specified, reserves a DID number. Cannot be used with '-Ext'
    [Parameter(ValueFromPipelineByPropertyName, ParameterSetName = 'Did')]
    [Parameter(ValueFromPipelineByPropertyName)]
    [switch] $Did,
    
    # When specified, reserves an Extension. Cannot be used with '-Did'
    [Parameter(ValueFromPipelineByPropertyName, ParameterSetName = 'Ext')]
    [Parameter(ValueFromPipelineByPropertyName)]
    [switch] $Ext
    
    # # When specified, will reserve a number that matches the ServiceProvider
    # [Parameter(ValueFromPipelineByPropertyName)]
    # [ValidateLength(1,255)]
    # [string] $ServiceProvider,
    
    # # When specified, will reserve a number that matches the ServiceId
    # [Parameter(ValueFromPipelineByPropertyName)]
    # [ValidateLength(1,255)]
    # [string] $ServiceId,
  )
  Begin{
    ConnectToServerIfNeeded
  }
  Process{
    try{
      # Assemble the query string based on the parameters specified
      if (($Did.IsPresent) -or (-not $Ext.IsPresent)){$Query = "Did=true&"}
      if (($Ext.IsPresent) -and (-not $Did.IsPresent)){$Query = "Ext=true&"}
      # if ($ServiceProvider){$Query += "ServiceProvider=$ServiceProvider&"}
      # if ($ServiceId){$Query += "ServiceId=$ServiceId&"}
      if ($Range){$Query += "Range=$Range&"}
      if ($ReservedMinutes){$Query += "ReservedMinutes=$ReservedMinutes&"}
    
      # What-If suport
      if (-not $(ShouldProcess $Query)) {
        return
      }

      # Assemble the URI
      [string] $Uri = $HTTP + $global:EzUdmTenant + '/' + $EzUdmRmNumberReservationApi + '?' + $Query
      Write-Verbose -Message $Uri
      $WebRequest = Invoke-WebRequest -Uri $Uri -Method Get -Credential $global:EzUdmCreds

      # Handle the JSON data returned
      # This code is just working theory right now, and should not be trusted
      $parsed = ConvertFrom-Json -InputObject $WebRequest.Content
      # $parsed.origin
      # $parsed.headers
      return $parsed
    }
    catch{
      Throw "An error has occured. The error was: '$_'."
    }
  }
  End{}
} # end function New-UdmRmNumberReservation
#endregion UdmRmNumberReservation

# helper functions
Function ShouldProcess {
  param(
    $Params,
    [string] $Operation
  )
  If ($PSCmdlet.MyInvocation.BoundParameters["WhatIf"].IsPresent) {
    # [string] $CommandName =  $PSCmdlet.MyInvocation.MyCommand.Name
    $stringParams = $Params
    if ($Params -and $Params.GetType().Name -eq 'HashTable') {
      $stringParams = ConvertTo-Json -InputObject $Params -Compress
    }
    return $PSCmdlet.ShouldProcess($stringParams)
  }
  return $True
}

Function ConnectToServerIfNeeded {
  # If the global variables EzUdmCreds or EzUdmTenant don't exist, there is no connection to the tenant. If that's the case, we need to connect.
  if ((-not $global:EzUdmCreds) -or (-not $global:EzUdmTenant)) {
    Write-Host 'You need to connect to UDM server before performing any actions.'
    Connect-UdmServer
  }
}
# end helper functions
